@@ -25,6 +25,7 @@
 #include <libzrtpcpp/ZrtpQueue.h>
 #include <libzrtpcpp/ZIDFile.h>
 #include <libzrtpcpp/ZRtp.h>
+#include <libzrtpcpp/ZrtpStateClass.h>
 #include <libzrtpcpp/ZrtpUserCallback.h>
 
 static TimeoutProvider<std::string, ost::ZrtpQueue*>* staticTimeoutProvider = NULL;
@@ -53,11 +54,6 @@ void ZrtpQueue::init()
     enableZrtp = false;
     started = false;
     zrtpEngine = NULL;
-
-    // senderCryptoContext = NULL;
-
-    // recvCryptoContext = NULL;
-
     senderZrtpSeqNo = 1;
 
     clientIdString = clientId;
@@ -66,7 +62,6 @@ void ZrtpQueue::init()
 
 ZrtpQueue::~ZrtpQueue() {
 
-//    cancelTimer();
     endQueue();
     stopZrtp();
 
@@ -74,18 +69,6 @@ ZrtpQueue::~ZrtpQueue() {
         delete zrtpUserCallback;
         zrtpUserCallback = NULL;
     }
-
-    /*
-    if (recvCryptoContext != NULL) {
-        delete recvCryptoContext;
-        recvCryptoContext = NULL;
-    }
-
-    if (senderCryptoContext != NULL) {
-        delete senderCryptoContext;
-        senderCryptoContext = NULL;
-    }
-    */
 }
 
 int32_t
@@ -242,6 +225,9 @@ ZrtpQueue::rtpDataPacket(IncomingRTPPkt* packet, int32 rtn,
                 return 0;
             }
         }
+        if (!zrtpEngine->inState(SecureState)) {
+            zrtpEngine->conf2AckSecure();
+        }
     }
 
     // virtual for profile-specific validation and processing.
@@ -492,9 +478,9 @@ void ZrtpQueue::srtpSecretsOn(std::string c, std::string s, bool verified)
 
   if (zrtpUserCallback != NULL) {
     zrtpUserCallback->secureOn(c);
-  }
-  if (zrtpUserCallback != NULL) {
-    zrtpUserCallback->showSAS(s, verified);
+    if (!s.empty()) {
+        zrtpUserCallback->showSAS(s, verified);
+    }
   }
 }
 
@@ -676,6 +662,15 @@ void ZrtpQueue::setPBXEnrollment(bool yesNo) {
 	zrtpEngine->setPBXEnrollment(yesNo);
 }
 
+
+int32 ZrtpQueue::getZid(uint8* data) {
+    if (data == NULL)
+        return 0;
+
+    if (zrtpEngine != NULL)
+        return zrtpEngine->getZid(data);
+}
+
 IncomingZRTPPkt::IncomingZRTPPkt(const unsigned char* const block, size_t len) :
         IncomingRTPPkt(block,len) {
 }
